/// TapeView - Recover data from MTX Audio Tapes
// For compilers that support precompilation, includes "wx/wx.h".
#include <wx/wxprec.h>
#ifndef WX_PRECOMP
#include <wx/wx.h>
#endif
#include <wx/valnum.h>
#include <wx/valgen.h>
#include <wx/spinctrl.h>
#include <vector>
#include <set>

enum Level { lvlInit, lvlLow, lvlHigh };

#define P_SHORT         0x00
#define P_LONG          0x01
#define P_ZERO          0x02
#define P_ONE           0x03
#define P_MANUAL        0x04
#define P_LEAD          0x10
#define P_MARK          0x20
#define P_DATA          0x30
#define M_VALID         0x02
#define M_TYPE          0x03
#define M_MARK          0xF0

class Pulse
   {
public:
   static unsigned int  nShort;
   static unsigned int  nMid;
   static unsigned int  nLong;
   Level        lvl;
   unsigned int flags;
   unsigned int n0;
   unsigned int n1;
   unsigned int n2;
   Pulse ()
      {
      lvl = lvlInit;
      flags = 0;
      n0 = 0;
      n1 = 0;
      n2 = 0;
      }
   Pulse (Level lvlNew, unsigned int n0New, unsigned int n1New, unsigned int n2New)
      {
      lvl = lvlNew;
      n0 = n0New;
      n1 = n1New;
      n2 = n2New;
      flags = Type (n2 - n0);
      }
   Pulse (const Pulse &p)
      {
      lvl = p.lvl;
      flags = p.flags;
      n0 = p.n0;
      n1 = p.n1;
      n2 = p.n2;
      }
   static unsigned int Type (unsigned int n)
      {
      if ( n < nShort ) return P_SHORT;
      else if ( n < nMid ) return P_ZERO;
      else if ( n < nLong ) return P_ONE;
      return P_LONG;
      }
   static void Baud (int nBaud, int nRate)
      {
      nShort = nRate / nBaud / 2;
      nMid = 3 * nRate / nBaud / 2;
      nLong = 5 * nRate / nBaud / 2;
      // printf ("Baud: Short = %d, Mid = %d, Long = %d.\n", nShort, nMid, nLong);
      }
   };

unsigned int Pulse::nShort = 0;
unsigned int Pulse::nMid = 0;
unsigned int Pulse::nLong = 0;

#define SV_BASTOP   0xFAA7      // Top of BASIC program
#define SV_NBTOP    0xFAA4      // Top of NODDY pages

enum BlockFmt {fmtUnknown, fmtBinary, fmtName, fmtSysVar, fmtBasic, fmtVars, fmtAsm};

class Tape;     // Forward definition

class Block
   {
public:
   static wxString sBlockFmt[];
   Tape *tape;
   unsigned int nBlock;
   bool bFound;
   bool bValid;
   enum BlockFmt fmt;
   unsigned int nAddr;
   unsigned int nLen;
   unsigned int nPulse;
   unsigned int nBit;
   unsigned char dat0;
   std::vector<unsigned char>   data;
   wxString sList;
   Block (Tape *t, unsigned int n);
   void Clear (void);
   bool AddBit (unsigned int iPulse, int iType);
   bool IsValid (unsigned int iAddr, unsigned int iLen = 1);
   unsigned int Byte (unsigned int iAddr);
   unsigned int Word (unsigned int iAddr);
   void ListHex (unsigned int iAddr);
   void ListSysVar (void);
   wxString OpCode (unsigned int &iAddr);
   void ListBasic (void);
   void ListAsm (void);
   void ListCode (void);
   void List (wxTextCtrl *txOut);
   };

enum WaveFmt { fmtUnk, fmtW08, fmtW16, fmtW32 };
enum ScanMode { scnValue, scnSlope, scnAccel, scnAvg3, scnAvg5, scnSlp5,
                scnHiLow, scnHLSlp, scnHLAcc, scnHLA3, scnHLA5, scnHLS5,
                scnCount};

class Tape
   {
public:
   static wxString sScanMode[];
   wxString sSource;
   std::vector<float> samples;
   std::vector<Pulse> pulses;
   std::vector<Block> blocks;
   std::set<unsigned int> split;
   std::set<unsigned int> merge;
   float fUp;                   /* Upwards threshold crossing */
   float fDn;                   /* Downwards threshold crossing */
   unsigned int nMinLead;       /* Minimum length of block leader */
   FILE *	pfTape;		/* File stream */
   enum WaveFmt fmt;		/* File format */
   unsigned int nChan;		/* Number of channels in WAV file */
   unsigned int	nRate;		/* WAV file sample rate (samples per second) */
   unsigned int	nBaud;		/* Baud rate (cycles per second) */
   unsigned int	nSamp;		/* Number of samples remaining */
   enum ScanMode scn;           /* Scan method */
   unsigned int nBlock;         /* Number of blocks found in most recent scan */
   char         sMsg[256];      /* Error message */
   unsigned int nFstPulse;      /* First pulse to display */
   float        fZoom;          /* Pixels per sample */
   float        sclSlope;       /* Scaling for slopes */
   float        sclAccel;       /* Scaling for accelerations */
   bool         bLockPulses;
   bool         bShowByte;      /* Highlight pulses forming a byte */
   unsigned int nShowPulse;     /* First pulse of byte to highlight */
   Tape (void) { Clear (); }
   void Clear (void);
   bool Fatal (const char *fmt, ...);
   bool Open (const char *psFile);
   bool Load (void);
   void Scan (void);
   float Test (enum ScanMode scn1, unsigned int iSamp);
   void SaveWave (const char *psFile);
   void SaveMTX (const char *psFile);
   void ListCode (const char *psFile);
   unsigned int SplitPulse (unsigned int iPulse, unsigned int iSamp);
   unsigned int MergePulse (unsigned int iPulse, unsigned int iSamp);
   bool LoadManual (const char *psFile);
   void SaveManual (const char *psFile);
   void ApplyManual (void);
   unsigned int SysWord (unsigned int iBlock, unsigned int iAddr);
   };

wxString Block::sBlockFmt[] = {
      "Unknown",
      "Binary",
      "Name",
      "SysVars",
      "Basic",
      "Vars",
      "Asm"
      };

Block::Block (Tape *t, unsigned int n)
   {
   tape = t;
   nBlock = n;
   Clear ();
   fmt = fmtUnknown;
   nLen = 0;
   nAddr = 0;
   }

void Block::Clear (void)
   {
   bFound = false;
   bValid = false;
   dat0 = 0;
   nBit = 0;
   nPulse = 0;
   data.clear ();
   sList = "";
   }

bool Block::AddBit (unsigned int iPulse, int iType)
   {
   if ( ! bFound )
      {
      bFound = true;
      bValid = true;
      nPulse = iPulse;
      }
   iType &= M_TYPE;
   if ( ( iType == P_SHORT ) || ( iType == P_LONG ) ) bValid = false;
   dat0 >>= 1;
   if ( ( iType == P_ONE ) || ( iType == P_LONG ) ) dat0 |= 0x80;
   if ( ++nBit >= 8 )
      {
      data.push_back (dat0);
      dat0 = 0;
      nBit = 0;
      if ( ( nLen > 0 ) && ( data.size () >= nLen ) ) return true;
      }
   return ( ( fmt == fmtUnknown ) && ( ! bValid ) );
   }

bool Block::IsValid (unsigned int iAddr, unsigned int iLen)
   {
   return ( ( iAddr >= nAddr ) && ( ( iAddr + iLen ) <= ( nAddr + data.size () ) ) );
   }

unsigned int Block::Byte (unsigned int iAddr)
   {
   if ( ( iAddr >= nAddr ) && ( iAddr < nAddr + data.size () ) ) return data[iAddr - nAddr];
   return 0;
   }

unsigned int Block::Word (unsigned int iAddr)
   {
   return ( Byte (iAddr + 1) << 8 ) | Byte (iAddr);
   }

void Block::ListHex (unsigned int iAddrSta)
   {
   wxString sRow1;
   wxString sRow2;
   unsigned int nAddrEnd = nAddr + nLen;
   unsigned int iCol;
   unsigned int iAddr = iAddrSta & 0xFFF0;
   if ( ( nLen == 0 ) || ( data.size () < nLen ) ) nAddrEnd = nAddr + data.size ();
   while ( iAddr < nAddrEnd )
      {
      sRow1.Printf ("0x%04X:", iAddr);
      sRow2 = " ";
      for ( iCol = 0; iCol < 16; ++iCol )
         {
         if ( ( iAddr >= iAddrSta ) && ( iAddr < nAddrEnd ) )
            {
            unsigned char by = data[iAddr - nAddr];
            sRow1 += wxString::Format (" %02X", by);
            if ( ( by >= 0x20 ) && ( by < 0x7F ) ) sRow2 += wxString::Format ("%c", by);
            else sRow2 += ".";
            }
         else
            {
            sRow1 += "   ";
            sRow2 += " ";
            }
         ++iAddr;
         }
      sList += sRow1 + sRow2 + "\n";
      }
   }

enum VarType { T_BYTE, T_WORD, T_JUMP, T_PGAD, T_VADR };

void Block::ListSysVar (void)
   {
   static const struct SysVar
      {
      unsigned int nAddr;
      unsigned int nLen;
      const char * psName;
      enum VarType type;
      } SysVars[] =
           {
           { 0xFA52, 10, "CTRBADR0", T_BYTE },
           { 0xFA5C, 10, "CTRBADR1", T_BYTE },
           { 0xFA66, 10, "CTRBADR2", T_BYTE },
           { 0xFA70, 10, "CTRBADR3", T_BYTE },
           { 0xFA7A, 1, "LSTPG", T_BYTE },
           { 0xFA7B, 2, "VARNAM", T_WORD },
           { 0xFA7D, 2, "VALBOT", T_WORD },
           { 0xFA7F, 2, "CALCBOT", T_WORD },
           { 0xFA81, 2, "CALCST", T_WORD },
           { 0xFA83, 2, "KBDBUF", T_WORD },
           { 0xFA85, 4, "USYNT", T_BYTE },
           { 0xFA89, 3, "USER", T_JUMP },
           { 0xFA8C, 3, "", T_JUMP },
           { 0xFA8F, 1, "IOPL", T_BYTE },
           { 0xFA90, 1,"REALBY", T_BYTE },
           { 0xFA91, 1, "KBFLAG", T_BYTE },
           { 0xFA92, 2, "STKLIM", T_WORD },
           { 0xFA94, 2, "SYSTOP", T_WORD },
           { 0xFA96, 2, "SSTACK", T_WORD },
           { 0xFA98, 3, "USERINT", T_JUMP },
           { 0xFA9B, 3, "NODLOC", T_JUMP },
           { 0xFA9E, 3, "FEXPAND", T_JUMP },
           { 0xFAA1, 3, "USERNOD", T_JUMP },
           { 0xFAA4, 3, "NBTOP", T_PGAD },
           { 0xFAA7, 3, "BASTOP", T_PGAD },
           { 0xFAAA, 2, "BASBOT", T_WORD },
           { 0xFAAC, 32, "BASTPG", T_WORD },
           { 0xFACC, 3, "ARRTOP", T_VADR },
           { 0xFACF, 2, "BASELN", T_WORD },
           { 0xFAD1, 1, "BASLNP", T_BYTE },
           { 0xFAD2, 1, "PAGE", T_BYTE },
           { 0xFAD3, 1, "CRNTPG", T_BYTE },
           { 0xFAD4, 1, "PGN1", T_BYTE },
           { 0xFAD5, 1, "PGN2", T_BYTE },
           { 0xFAD6, 2, "PGTOP", T_WORD },
           { 0xFAD8, 105, "GOSTK", T_PGAD },
           { 0xFB41, 2, "GOPTR", T_WORD },
           { 0xFB43, 1, "GOSNUM", T_BYTE },
           { 0xFB44, 1, "FORCOUNT", T_BYTE },
           { 0xFB45, 1, "CTYSLT", T_BYTE },
           { 0xFB46, 3, "DATAAD", T_PGAD },
           { 0xFB49, 2, "DESAVE", T_WORD },
           { 0xFB4B, 509, "START", T_BYTE },
           { 0xFD48, 3, "SETCALL", T_JUMP },
           { 0xFD4B, 3, "RICHJL", T_BYTE },
           { 0xFD4E, 3, "USRRST", T_JUMP },
           { 0xFD51, 3, "USERIO", T_JUMP },
           { 0xFD54, 3, "USERROR", T_JUMP },
           { 0xFD57, 7, "CLOCK", T_BYTE },
           { 0xFD5E, 1, "INTFFF", T_BYTE },
           { 0xFD5F, 1, "CASBAUD", T_BYTE },
           { 0xFD60, 1, "MIDVAL", T_BYTE },
           { 0xFD61, 4, "RETSAVE", T_WORD },
           { 0xFD65, 2, "VAZERO", T_WORD },
           { 0xFD67, 1, "VERIF", T_BYTE },
           { 0xFD68, 1, "TYPE", T_BYTE }
           };
   unsigned int nSysVar = sizeof (SysVars) / sizeof (SysVars[0]);
   unsigned int iSysVar = 0;
   wxString sRow;
   unsigned int nAddrEnd = nAddr + nLen;
   unsigned int iAddr = nAddr;
   unsigned int nByte = 0;
   while ( ( iSysVar < nSysVar -1 ) && ( SysVars[iSysVar].nAddr < iAddr ) ) ++iSysVar;
   if ( ( nLen == 0 ) || ( data.size () < nLen ) ) nAddrEnd = nAddr + data.size ();
   while ( iAddr < nAddrEnd )
      {
      if ( iAddr == SysVars[iSysVar].nAddr )
         {
         sRow.Printf ("0x%04X %s:", iAddr, SysVars[iSysVar].psName);
         if ( iSysVar < nSysVar - 1 ) ++iSysVar;
         }
      else if ( sRow.IsEmpty () )
         {
         sRow.Printf ("0x%04X:", iAddr);
         }
      sRow += wxString::Format (" %02X", data[iAddr - nAddr]);
      ++iAddr;
      if ( ( iAddr == SysVars[iSysVar].nAddr ) || ( ++nByte >= 16 ) )
         {
         sList += sRow + "\n";
         sRow.Clear ();
         nByte = 0;
         }
      }
   if ( ! sRow.IsEmpty () ) sList += sRow + "\n";
   }

wxString Block::OpCode (unsigned int &iAddr)
   {
#include "mnem.cpp"
   wxString sCode;
   const char *psOpcode =  NULL;
   const char *psIXReg[] = { "HL", "IX", "IY" };
   int iIndex = 0;
   unsigned char b;
   unsigned int w;
   sCode = wxString::Format ("%04X\t", iAddr);
   b	= Byte (iAddr++);
   if ( b == 0xDD )
      {
      iIndex   =  1;
      b	= Byte (iAddr++);
      }
   else if ( b == 0xFD )
      {
      iIndex   =  2;
      b	= Byte (iAddr++);
      }
   switch (b)
      {
      case  0xCB:
      {
      b	= Byte (iAddr++);
      psOpcode =  psMnemCB[b];
      break;
      }
      case  0xED:
      {
      b	= Byte (iAddr++);
      if ( ( b >= 0x40 ) && ( b < 0x7C ) )      psOpcode =  psMnemED[b - 0x40];
      else if ( ( b >= 0xA0 ) && ( b < 0xBC ) ) psOpcode =  psMnemED[b - 0xA0 + 0x3C];
      break;
      }
      default:
      {
      psOpcode =  psMnem[b];
      }
      }
   if ( ( psOpcode == NULL ) || ( *psOpcode == '*' ) )
      {
      sCode += wxString::Format ("DB\t0x%02X\n", b);
      return   sCode;
      }
   while ( *psOpcode )
      {
      switch (*psOpcode)
         {
         case  '&':
         case  '^':
         {
         b	= Byte (iAddr++);
         sCode += wxString::Format ("0x%02X", b);
         break;
         }
         case  '?':
         case  '#':
         case  '~':
         case  '!':
         {
         sCode += wxString::Format ("0x%04X", Word (iAddr));
         iAddr += 2;
         break;
         }
         case  '$':
         {
         b	= Byte (iAddr++);
         w  	= iAddr + b;
         if ( b & 0x80 )   w  -= 0x100;
         sCode += wxString::Format ("0x%04X", w);
         break;
         }
         case  ':':
         {
         sCode += wxString::Format ("0x%02X", b - 0xC7);
         break;
         }
         case  '%':
         {
         sCode += wxString::Format ("%s", psIXReg[iIndex]);
         break;
         }
         case  '@':
         {
         if ( iIndex == 0 )
            {
            sCode += wxString::Format ("(HL)");
            }
         else
            {
            sCode += wxString::Format ("(%s+0x%02x)", psIXReg[iIndex], Byte (iAddr++));
            }
         break;
         }
         default:
         {
         sCode += wxString::Format ("%c", *psOpcode);
         break;
         }
         }
      ++psOpcode;
      }
   sCode += "\n";
   return sCode;
   }

void Block::ListBasic (void)
   {
   static const char * psTokens[] =
      {
      "REM",			// 128
      "CLS",			// 129
      "ASSEM",		        // 130
      "AUTO",			// 131
      "BAUD",			// 132
      "VS",			// 133
      "CONT",			// 134
      "USER",			// 135
      "CRVS",			// 136
      "CLEAR",		        // 137
      "CLOCK",		        // 138
      "ATTR",			// 139
      "COLOUR",		        // 140
      "INK",			// 141
      "CSR",			// 142
      "DATA",			// 143
      "PRINT",		        // 144
      "DIM",			// 145
      "ADJSPR",		        // 146
      "EDIT",			// 147
      "NEXT",			// 148
      "FOR",			// 149
      "GOTO",			// 150
      "GOSUB",		        // 151
      "INPUT",		        // 152
      "IF",			// 153
      "MVSPR",		        // 154
      "LIST",			// 155
      "LET",			// 156
      "LLIST",		        // 157
      "LOAD",			// 158
      "LPRINT",		        // 159
      "SPRITE",		        // 160
      "CTLSPR",		        // 161
      "MODE",			// 162
      "NEW",			// 163
      "PAPER",		        // 164
      "NODDY",		        // 165
      "ON",			// 166
      "OUT",			// 167
      "PLOD",			// 168
      "PANEL",		        // 169
      "GENPAT",		        // 170
      "PAUSE",		        // 171
      "PHI",			// 172
      "POKE",			// 173
      "RAND",			// 174
      "RETURN",		        // 175
      "READ",			// 176
      "VIEW",			// 177
      "RESTORE",		// 178
      "ROM",			// 179
      "RUN",			// 180
      "SAVE",			// 181
      "SOUND",		        // 182
      "EDITOR",		        // 183
      "DSI",			// 184
      "PLOT",			// 185
      "STOP",			// 186
      "ANGLE",		        // 187
      "SBUF",			// 188
      "VERIFY",		        // 189
      "DRAW",			// 190
      "ARC",			// 191
      "CIRCLE",		        // 192
      "LINE",			// 193
      "CODE",			// 194
      "ELSE",		// " ELSE",			// 195
      "FK",			// 196
      "OFF",			// 197
      "STEP",		// " STEP",			// 198
      "THEN",		// " THEN",			// 199
      "TO",		// " TO",			// 200
      "[C9]",			// 201
      "[CA]",			// 202
      "[CB]",			// 203
      "[CC]",			// 204
      "[CD]",			// 205
      "[CE]",			// 206
      "+",		// " +",			// 207
      "-",		// " -",			// 208
      "*",		// " *",			// 209
      "/",		// " /",			// 210
      "^",		// " ^",			// 211
      "=",		// " =",			// 212
      ">",		// " >",			// 213
      "<",		// " <",			// 214
      ">=",		// " >=",			// 215
      "<=",		// " <=",			// 216
      "<>",		// " <>",			// 217
      "AND",		// " AND",			// 218
      "OR",		// " OR",			// 219
      "NOT",		// " NOT",			// 220
      "ABS",			// 221
      "ATAN",			// 222
      "COS",			// 223
      "EXP",			// 224
      "FRE",			// 225
      "INT",			// 226
      "INT",			// 227
      "LN",			// 228
      "PEEK",			// 229
      "SGN",			// 230
      "SIN",			// 231
      "SQR",			// 232
      "TAN",			// 233
      "IMP",			// 234
      "USR",			// 235
      "LN",			// 236
      "ASC",			// 237
      "LEN",			// 238
      "VAL",			// 239
      "LN",			// 240
      "MOD",			// 241
      "PI",		   	// 242
      "RND",			// 243
      "PI",			// 244
      "CHR$",                   // 245
      "SPK$",                   // 246
      "INKEY$",                 // 247
      "LEFT$",                  // 248
      "MID$",                   // 249
      "RIGHT$",                 // 250
      "GR$",                    // 251
      "STR$",                   // 252
      "TIME$"                   // 253
      };
   wxString sRow;
   unsigned int nAddrEnd = nAddr + nLen;
   unsigned int iAddr = nAddr;
   unsigned int iBasTop = tape->SysWord (nBlock, SV_BASTOP);
   if ( ( nLen == 0 ) || ( data.size () < nLen ) ) nAddrEnd = nAddr + data.size ();
   if ( ( iBasTop >= 0x4000 ) && ( iBasTop < nAddrEnd ) ) nAddrEnd = iBasTop;
   while ( iAddr < nAddrEnd )
      {
      sList += wxString::Format ("0x%04X: ", iAddr);
      unsigned int nLineEnd = iAddr + Word (iAddr);
      if ( nLineEnd > nAddrEnd )
         {
         sList += wxString::Format ("Invalid line length: 0x%04X.\n", nLineEnd - iAddr);
         break;
         }
      iAddr += 2;
      sList += wxString::Format ("%d ", Word (iAddr));
      iAddr += 2;
      bool bQuote = false;
      while ( iAddr < nLineEnd )
         {
         unsigned int ch = Byte (iAddr++);
         if ( ch == 194 )
            {
            // Code
            unsigned int nAsmEnd = iAddr + Word (iAddr) + 2;
            iAddr += 2;
            sList += "CODE\n";
            while ( iAddr < nAsmEnd )
               {
               sList += OpCode (iAddr);
               }
            sList += "CODE Labels (yet to be decoded)\n";
            if ( ( iAddr & 0x0F ) != 0 ) sList += wxString::Format ("%04X\t", iAddr);
            while ( iAddr < nLineEnd )
               {
               if ( ( iAddr & 0x0F ) == 0 ) sList += wxString::Format ("\n%04X\t", iAddr);
               ch = Byte (iAddr++);
               if ( ( ch >= 0x20 ) && ( ch < 0x7F ) )	sList += wxString::Format ("  %c", ch);
               else sList += wxString::Format (" %02X", ch);
               }
            sList += "\nEnd of CODE statement.\n";
            }
         else if ( ( ch >= 0x20 ) && ( ch < 0x7F ) )
            {
            sList += wxString::Format ("%c", ch);
            }
         else if ( ( ! bQuote ) && ( ch >= 1 ) && ( ch <= 26 ) )
            {
            sList += wxString::Format ("%c", ch + 0x40);
            }
         else if ( ( ! bQuote ) && ( ch >= 128 ) && ( ch <= 253 ) )
            {
            int lch = sList.Last ();
            if ( ( lch != ' ' ) && ( lch != '\t' ) ) sList += ' ';
            sList += psTokens[ch-128];
            sList += " ";
            }
         else
            {
            sList += wxString::Format ("[%02X]", ch);
            }
         if ( ch == '"' ) bQuote = ! bQuote;
         }
      sList += "\n";
      }
   nAddrEnd = nAddr + nLen;
   unsigned int iNodTop = tape->SysWord (nBlock, SV_NBTOP);
   if ( ( nLen == 0 ) || ( data.size () < nLen ) ) nAddrEnd = nAddr + data.size ();
   if ( ( iNodTop >= 0x4000 ) && ( iNodTop < nAddrEnd ) ) nAddrEnd = iNodTop;
   if ( iAddr < nAddrEnd )
       {
       sList += "NODDY pages:\n";
       ListHex (iAddr);
       }
   }

void Block::ListAsm (void)
   {
   unsigned int nAddrEnd = nAddr + nLen;
   unsigned int iAddr = nAddr;
   if ( ( nLen == 0 ) || ( data.size () < nLen ) ) nAddrEnd = nAddr + data.size ();
   while ( iAddr <= nAddrEnd )
      {
      sList += OpCode (iAddr);
      }
   }

void Block::ListCode (void)
   {
   if ( sList.IsEmpty () )
      {
      switch (fmt)
         {
         case fmtUnknown:
         case fmtBinary:
         case fmtName:
         case fmtVars:
            ListHex (nAddr);
            break;
         case fmtSysVar:
            ListSysVar ();
            break;
         case fmtBasic:
            ListBasic ();
            break;
         case fmtAsm:
            ListAsm ();
            break;
         }
      }
   }

void Block::List (wxTextCtrl *txOut)
   {
   txOut->Freeze ();
   if ( ! bFound )
      {
      txOut->SetValue ("Not Found !\n");
      }
   else
      {
      ListCode ();
      txOut->SetValue (sList);
      }
   txOut->Thaw ();
   }

wxString Tape::sScanMode[] = {
   "Values",
   "Slopes",
   "Accels",
   "3 Avgs",
   "5 Avgs",
   "Slope 5",
   "High-Low",
   "H-L Slope",
   "H-L Accel",
   "H-L 3 Avgs",
   "H-L 5 Avgs",
   "H-L Slope 5"
   };

void Tape::Clear (void)
   {
   pulses.clear ();
   samples.clear ();
   blocks.clear ();
   split.clear ();
   merge.clear ();
   fUp = 0.2;
   fDn = 0.2;
   nMinLead = 256;
   nBaud = 2400;
   sMsg[0] = '\0';
   nFstPulse = 0;
   fZoom = 1.0;
   nSamp = 0;
   nBlock = 0;
   scn = scnValue;
   bLockPulses = false;
   }

bool Tape::Fatal (const char *fmt, ...)
   {
   va_list	vars;
   va_start (vars, fmt);
   vsnprintf (sMsg, sizeof (sMsg), fmt, vars);
   va_end (vars);
   return false;
   }

bool Tape::Open (const char *psFile)
   {
   unsigned int nLen;
   char sID[5];
   fmt = fmtUnk;
   pfTape = fopen (psFile, "rb");
   if ( pfTape == NULL ) return Fatal ("Failed to open input file %s", psFile);
   fread (sID, sizeof (char), 4, pfTape);
   sID[4] = '\0';
   if ( strcmp (sID, "RIFF") )	return Fatal ("Tape WAV file does not have a RIFF header");
   fread (&nLen, sizeof (nLen), 1, pfTape);
   fread (sID, sizeof (char), 4, pfTape);
   if ( strcmp (sID, "WAVE") ) return Fatal ("Tape WAV file is not WAVE format");
   while ( fread (sID, sizeof (char), 4, pfTape) == 4 )
      {
      fread (&nLen, sizeof (nLen), 1, pfTape);
      if ( ! strcmp (sID, "fmt ") )
         {
         unsigned short int wFmt;
         unsigned short int wChan;
         unsigned int nBPS;
         unsigned short int wAlign;
         unsigned int nBits;
         unsigned short int wBits;
         fread (&wFmt, sizeof (wFmt), 1, pfTape);
         if ( wFmt != 1 ) return Fatal ("Tape WAV file is compressed");
         fread (&wChan, sizeof (wChan), 1, pfTape);
         nChan = wChan;
         fread (&nRate, sizeof (nRate), 1, pfTape);
         fread (&nBPS, sizeof (nBPS), 1, pfTape);
         fread (&wAlign, sizeof (wAlign), 1, pfTape);
         if ( nLen == 16 )
            {
            fread (&wBits, sizeof (wBits), 1, pfTape);
            nBits = wBits;
            }
         else if ( nLen == 18 )
            {
            fread (&nBits, sizeof (nBits), 1, pfTape);
            }
         else
            {
            return Fatal ("Tape WAV file has unexpected fmt chunk length");
            }
         if ( nBits == 8 )       fmt = fmtW08;
         else if ( nBits == 16 ) fmt = fmtW16;
         else if ( nBits == 32 ) fmt = fmtW32;
         else return Fatal ("Tape WAV file has unsupported sample length");
         }
      else if ( ! strcmp (sID, "data") )
         {
         if ( fmt == fmtUnk )      return Fatal ("No fmt chunk in WAV file before data");
         Clear ();
         sSource = psFile;
         if ( fmt == fmtW08 ) nSamp = nLen / nChan;
         else if ( fmt == fmtW16 ) nSamp = nLen / nChan / 2;
         else if ( fmt == fmtW32 ) nSamp = nLen / nChan / 4;
         return true;
         }
      else
         {
         fseek (pfTape, nLen, SEEK_CUR);
         }
      }
   return Fatal ("No data chunk found in Tape WAV file");
   }

bool Tape::Load (void)
   {
   float fVal;
   float fMin;
   float fMax;
   float fMinA;
   float fMaxA;
   float fMean = 0;
   unsigned int iChan;
   unsigned int iSamp;
   unsigned int nRead;
   unsigned char by;
   short int wd;
   float f;
   for ( iSamp = 0; iSamp < nSamp; ++iSamp )
      {
      fVal = 0.0;
      for ( iChan = 0; iChan < nChan; ++iChan )
         {
         switch (fmt)
            {
            case fmtW08:
               nRead = fread (&by, sizeof (by), 1, pfTape);
               fVal += by / 128.0 - 1.0;
               break;
            case fmtW16:
               nRead = fread (&wd, sizeof (wd), 1, pfTape);
               fVal += wd / 32768.0;
               break;
            case fmtW32:
               nRead = fread (&f, sizeof (f), 1, pfTape);
               fVal += f;
               break;
            }
         if ( nRead == 0 )
            {
            return Fatal ("Premature end of file");
            }
         }
      if ( iSamp == 0 )
         {
         fMin = fVal;
         fMax = fVal;
         }
      else if ( fVal < fMin )
         {
         fMin = fVal;
         }
      else if ( fVal > fMax )
         {
         fMax = fVal;
         }
      samples.push_back (fVal);
      }
   fMean /= nSamp;
   fMax -= fMean;
   fMin -= fMean;
   if ( -fMin > fMax ) fMax = - fMin;
   for ( iSamp = 0; iSamp < nSamp; ++iSamp )
      {
      samples[iSamp] = ( samples[iSamp] - fMean ) / fMax;
      }
   sclSlope = 1.0;
   sclAccel = 1.0;
   fMin = Test (scnSlope, 0);
   fMax = fMin;
   fMinA = Test (scnAccel, 0);
   fMaxA = fMinA;
   for ( iSamp = 1; iSamp < nSamp; ++iSamp )
      {
      fVal = Test (scnSlope, iSamp);
      if ( fVal < fMin ) fMin = fVal;
      else if ( fVal > fMax ) fMax = fVal;
      fVal = Test (scnAccel, iSamp);
      if ( fVal < fMinA ) fMinA = fVal;
      else if ( fVal > fMaxA ) fMaxA = fVal;
      }
   sclSlope = ( fMax > - fMin ) ? fMax : - fMin;
   sclAccel = ( fMaxA > - fMinA ) ? fMaxA : - fMinA;
   Pulse::Baud (nBaud, nRate);
   nFstPulse = 0;
   fZoom = 1.0;
   fUp = 0.2;
   fDn = -0.2;
   return true;
   }

float Tape::Test (enum ScanMode scn1, unsigned int iSamp)
   {
   switch (scn1)
      {
      case scnValue:
      case scnHiLow:
         return samples[iSamp];
      case scnSlope:
      case scnHLSlp:
         if ( iSamp == 0 )
            return 2.0 * ( samples[1] - samples[0] ) / sclSlope;
         else if ( iSamp == nSamp - 1 )
            return 2.0 * ( samples[nSamp-1] - samples[nSamp-2] ) / sclSlope;
         return ( samples[iSamp+1] - samples[iSamp-1] ) / sclSlope;
      case scnAccel:
      case scnHLAcc:
         if ( iSamp == 0 )
            return ( samples[2] - 2.0 * samples[1] + samples[0] ) / sclAccel;
         else if ( iSamp == nSamp - 1 )
            return ( samples[nSamp-1] - 2.0 * samples[nSamp-2] + samples[nSamp-3] ) / sclAccel;
         return ( samples[iSamp+1] - 2.0 * samples[iSamp] + samples[iSamp-1] ) / sclAccel;
      case scnAvg3:
      case scnHLA3:
         if ( iSamp == 0 )
            return ( samples[1] + samples[0] ) / 2.0;
         else if ( iSamp == nSamp - 1 )
            return ( samples[nSamp-1] + samples[nSamp-2] ) / 2.0;
         return ( samples[iSamp+1] + samples[iSamp] + samples[iSamp-1] ) / 3.0;
      case scnAvg5:
      case scnHLA5:
         if ( iSamp == 0 )
            return ( samples[2] + samples[1] + samples[0] ) / 3.0;
         if ( iSamp == 1 )
            return ( samples[3] + samples[2] + samples[1] + samples[0] ) / 4.0;
         else if ( iSamp == nSamp - 1 )
            return ( samples[nSamp-1] + samples[nSamp-2] + samples[nSamp-3] ) / 3.0;
         else if ( iSamp == nSamp - 2 )
            return ( samples[nSamp-1] + samples[nSamp-2] + samples[nSamp-3] + samples[nSamp-4] ) / 4.0;
         return ( samples[iSamp+2] + samples[iSamp+1] + samples[iSamp]
            + samples[iSamp-1] + samples[iSamp-2] ) / 5.0;
      case scnSlp5:
      case scnHLS5:
         if ( iSamp == 0 )
            return ( samples[2] - samples[0] ) / sclSlope;
         if ( iSamp == 1 )
            return ( samples[3] + samples[2] - samples[1] - samples[0] ) / sclSlope / 2.0;
         else if ( iSamp == nSamp - 1 )
            return ( samples[nSamp-1] - samples[nSamp-3] ) / sclSlope;
         else if ( iSamp == nSamp - 2 )
            return ( samples[nSamp-1] + samples[nSamp-2] - samples[nSamp-3]
               - samples[nSamp-4] ) / sclSlope / 2.0;
         return ( samples[iSamp+2] + samples[iSamp+1]
            - samples[iSamp-1] - samples[iSamp-2] ) / sclSlope / 3.0;
      }
   }

void Tape::Scan (void)
   {
   wxBusyCursor wait;
   enum Level lvlPrev = lvlInit;
   enum Level lvlCurr = lvlInit;
   enum Level lvlMark;
   enum { stWait, stLead, stData } state = stWait;
   unsigned int n0 = 0;
   unsigned int n1 = 0;
   unsigned int n2 = 0;
   unsigned int nPSL = 0;
   unsigned int iBlock = 0;
   // printf ("Scan: Short = %d, Mid = %d, Long = %d.\n", Pulse::nShort, Pulse::nMid, Pulse::nLong);
   if ( bLockPulses )
      {
      for (unsigned int iPulse = 0; iPulse < pulses.size (); ++iPulse)
         {
         pulses[iPulse].flags &= ( M_TYPE | P_MANUAL );
         }
      }
   else
      {
      pulses.clear ();
      if ( scn < scnHiLow )
         {
         for ( n2 = 0; n2 < nSamp; ++n2 )
            {
            float fVal = Test (scn, n2);
            if ( fVal > fUp )      lvlCurr = lvlHigh;
            else if ( fVal < fDn ) lvlCurr = lvlLow;
            if ( lvlCurr != lvlPrev )
               {
               pulses.push_back (Pulse (lvlPrev, n0, n1, n2));
               n0 = n1;
               n1 = n2;
               }
            lvlPrev = lvlCurr;
            }
         }
      else
         {
         float fMin = Test (scn, 0);
         float fMax = fMin;
         float fVal = Test (scn, 1);
         if ( fVal > fMax )
            {
            lvlCurr = lvlHigh;
            fMax = fVal;
            n2 = 1;
            }
         else
            {
            lvlCurr = lvlLow;
            fMin = fVal;
            n2 = 1;
            }
         for ( unsigned int iSamp = 2; iSamp < nSamp; ++iSamp )
            {
            fVal = Test (scn, iSamp);
            if ( lvlCurr == lvlHigh )
               {
               if ( fVal > fMax )
                  {
                  fMax = fVal;
                  n2 = iSamp;
                  }
               else if ( fVal < fMax + 2.0 * fDn )
                  {
                  pulses.push_back (Pulse (lvlCurr, n0, n1, n2));
                  n0 = n1;
                  n1 = n2;
                  n2 = iSamp;
                  lvlCurr = lvlLow;
                  fMin = fVal;
                  }
               }
            else
               {
               if ( fVal < fMin )
                  {
                  fMin = fVal;
                  n2 = iSamp;
                  }
               else if ( fVal > fMin + 2.0 * fUp )
                  {
                  pulses.push_back (Pulse (lvlCurr, n0, n1, n2));
                  n0 = n1;
                  n1 = n2;
                  n2 = iSamp;
                  lvlCurr = lvlHigh;
                  fMax = fVal;
                  }
               }
            }
         }
      if ( pulses.back ().n2 < nSamp - 1 )
         {
         pulses.push_back (Pulse (lvlCurr, n0, n1, nSamp - 1));
         }
      }
   nBlock = 0;
   for (unsigned int iPulse = 0; iPulse < pulses.size (); ++iPulse)
      {
      Pulse &pulse = pulses[iPulse];
      if ( state == stWait )
         {
         if ( pulse.flags == P_ZERO )
            {
            state = stLead;
            nPSL = iPulse;
            }
         }
      else if ( state == stLead )
         {
         if ( ( pulse.flags == P_ONE ) && ( iPulse - nPSL > nMinLead ) )
            {
            lvlMark = pulse.lvl;
            pulses[nPSL].flags |= P_LEAD;
            pulses[iPulse].flags |= P_MARK;
            state = stData;
            iBlock = nBlock;
            if ( ++nBlock > blocks.size () ) blocks.push_back (Block (this, iBlock));
            else blocks[iBlock].Clear ();
            if ( blocks[iBlock].fmt == fmtUnknown )
               {
               if ( iBlock == 0 )
                  {
                  blocks[iBlock].fmt = fmtName;
                  blocks[iBlock].nLen = 18;
                  }
               else if ( blocks[iBlock - 1].fmt == fmtName )
                  {
                  if ( blocks[iBlock - 1].Byte (0) == 0xFF )
                     {
                     unsigned int iSysSta = blocks[iBlock - 1].Word (16);
                     if ( ( iSysSta >= 0xD000 ) && ( iSysSta < 0xFB4B ) )
                        {
                        blocks[iBlock].fmt = fmtSysVar;
                        blocks[iBlock].nAddr = iSysSta;
                        blocks[iBlock].nLen = 0xFB4B - iSysSta;
                        }
                     }
                  }
               else if ( blocks[iBlock - 1].fmt == fmtSysVar )
                  {
                  unsigned int iLen = blocks[iBlock - 1].Word (0xFACC);
                  if ( iLen < 0xB000 )
                     {
                     blocks[iBlock].fmt = fmtBasic;
                     blocks[iBlock].nAddr = 0x4000;
                     blocks[iBlock].nLen = iLen;
                     }
                  }
               else if ( ( iBlock >= 2 ) && ( blocks[iBlock - 2].fmt == fmtSysVar ) )
                  {
                  unsigned int iSta = blocks[iBlock - 2].Word (0xFA7B);
                  unsigned int iEnd = blocks[iBlock - 2].Word (0xFA81);
                  if ( ( iSta >= 0xC000 ) && ( iEnd >= iSta ) && ( iEnd < 0xF000 ) )
                     {
                     blocks[iBlock].fmt = fmtVars;
                     blocks[iBlock].nAddr = iSta;
                     blocks[iBlock].nLen = iEnd - iSta;
                     }
                  }
               }
            else if ( blocks[iBlock].fmt == fmtName )
               {
               blocks[iBlock].nLen = 18;
               }
            else if ( blocks[iBlock].fmt == fmtSysVar )
               {
               if ( ( iBlock > 0 ) && ( blocks[iBlock - 1].fmt == fmtName ) )
                  {
                  if ( blocks[iBlock - 1].Byte (0) == 0xFF )
                     {
                     unsigned int iSysSta = blocks[iBlock - 1].Word (16);
                     if ( ( iSysSta >= 0xD000 ) && ( iSysSta < 0xFB4B ) )
                        {
                        blocks[iBlock].nAddr = iSysSta;
                        blocks[iBlock].nLen = 0xFB4B - iSysSta;
                        }
                     }
                  }
               }
            else if ( blocks[iBlock].fmt == fmtBasic )
               {
               if ( ( iBlock > 0 ) && ( blocks[iBlock - 1].fmt == fmtSysVar ) )
                  {
                  unsigned int iLen = blocks[iBlock - 1].Word (0xFACC);
                  if ( iLen < 0xB000 )
                     {
                     blocks[iBlock].nAddr = 0x4000;
                     blocks[iBlock].nLen = iLen;
                     }
                  }
               }
            else if ( blocks[iBlock].fmt == fmtVars )
               {
               if ( ( iBlock >= 2 ) && ( blocks[iBlock - 2].fmt == fmtSysVar ) )
                  {
                  unsigned int iSta = blocks[iBlock - 2].Word (0xFA7B);
                  unsigned int iEnd = blocks[iBlock - 2].Word (0xFA81);
                  if ( iEnd >= iSta )
                     {
                     blocks[iBlock].nAddr = iSta;
                     blocks[iBlock].nLen = iEnd - iSta;
                     }
                  }
               }
            }
         else if ( pulse.flags != P_ZERO )
            {
            state = stWait;
            }
         }
      else // if ( state == stData )
         {
         if ( pulse.lvl == lvlMark )
            {
            if ( blocks[iBlock].nBit == 0 ) pulse.flags |= P_DATA;
            if ( blocks[iBlock].AddBit (iPulse, pulse.flags) ) state = stWait;
            }
         }
      }
   }

unsigned int Tape::SplitPulse (unsigned int iPulse, unsigned int iSamp)
   {
   // printf ("SplitPulse: iPulse = %d, iSamp = %d\n", iPulse, iSamp);
   while ( ( iPulse < pulses.size () ) && ( pulses[iPulse].n2 < iSamp ) ) ++iPulse;
   if ( pulses[iPulse].n1 == iSamp ) return iPulse;
   // printf ("iPulse = %d, n0 = %d, n1 = %d, n2 = %d\n", iPulse, pulses[iPulse].n0, pulses[iPulse].n1, pulses[iPulse].n2);
   // printf ("@1 %d: (%d, %d, %d) %d: (%d, %d, %d) %d: (%d, %d, %d)\n",
   //    iPulse-1, pulses[iPulse-1].n0, pulses[iPulse-1].n1, pulses[iPulse-1].n2, 
   //    iPulse, pulses[iPulse].n0, pulses[iPulse].n1, pulses[iPulse].n2, 
   //    iPulse+1, pulses[iPulse+1].n0, pulses[iPulse+1].n1, pulses[iPulse+1].n2);
   enum Level lvl = ( pulses[iPulse].lvl == lvlHigh ) ? lvlLow : lvlHigh;
   unsigned int n0 = pulses[iPulse].n1;
   unsigned int n2 = pulses[iPulse].n2;
   pulses[iPulse].n2 = iSamp;
   pulses[iPulse].flags = Pulse::Type (pulses[iPulse].n2 - pulses[iPulse].n0);
   // printf ("@2 %d: (%d, %d, %d) %d: (%d, %d, %d) %d: (%d, %d, %d)\n",
   //    iPulse-1, pulses[iPulse-1].n0, pulses[iPulse-1].n1, pulses[iPulse-1].n2, 
   //    iPulse, pulses[iPulse].n0, pulses[iPulse].n1, pulses[iPulse].n2, 
   //    iPulse+1, pulses[iPulse+1].n0, pulses[iPulse+1].n1, pulses[iPulse+1].n2);
   ++iPulse;
   pulses[iPulse].n0 = iSamp;
   pulses[iPulse].flags = Pulse::Type (pulses[iPulse].n2 - pulses[iPulse].n0);
   // printf ("@3 %d: (%d, %d, %d) %d: (%d, %d, %d) %d: (%d, %d, %d)\n",
   //    iPulse-1, pulses[iPulse-1].n0, pulses[iPulse-1].n1, pulses[iPulse-1].n2, 
   //    iPulse, pulses[iPulse].n0, pulses[iPulse].n1, pulses[iPulse].n2, 
   //    iPulse+1, pulses[iPulse+1].n0, pulses[iPulse+1].n1, pulses[iPulse+1].n2);
   pulses.insert (pulses.begin () + iPulse, Pulse (lvl, n0, iSamp, n2));
   pulses[iPulse].flags |= P_MANUAL;
   // printf ("@4 %d: (%d, %d, %d) %d: (%d, %d, %d) %d: (%d, %d, %d)\n",
   //    iPulse-1, pulses[iPulse-1].n0, pulses[iPulse-1].n1, pulses[iPulse-1].n2, 
   //    iPulse, pulses[iPulse].n0, pulses[iPulse].n1, pulses[iPulse].n2, 
   //    iPulse+1, pulses[iPulse+1].n0, pulses[iPulse+1].n1, pulses[iPulse+1].n2);
   return iPulse;
   }

unsigned int Tape::MergePulse (unsigned int iPulse, unsigned int iSamp)
   {
   while ( ( iPulse < pulses.size () )
      && ( ( pulses[iPulse].n1 + pulses[iPulse].n2 ) / 2 < iSamp ) ) ++iPulse;
   --iPulse;
   pulses[iPulse].n2 = pulses[iPulse + 1].n2;
   pulses[iPulse].flags = Pulse::Type (pulses[iPulse].n2 - pulses[iPulse].n0) | P_MANUAL;
   pulses.erase (pulses.begin () + (iPulse + 1));
   pulses[iPulse+1].n0 = pulses[iPulse].n1;
   pulses[iPulse+1].n1 = pulses[iPulse].n2;
   pulses[iPulse+1].flags = Pulse::Type (pulses[iPulse+1].n2 - pulses[iPulse+1].n0);
   pulses[iPulse+2].n0 = pulses[iPulse].n2;
   pulses[iPulse+2].flags = Pulse::Type (pulses[iPulse+2].n2 - pulses[iPulse+2].n0);
   return iPulse;
   }

void Tape::SaveWave (const char *psFile)
   {
   int iVal;
   unsigned short int wVal;
   unsigned int nChanOut = 1;
   // printf ("Open file %s\n", psFile);
   FILE * pfTapeOut = fopen (psFile, "wb");
   fwrite ("RIFF", sizeof (char), 4, pfTapeOut);
   unsigned int nSampOut = 0;
   fwrite (&nSampOut, sizeof (nSampOut), 1, pfTapeOut);
   fwrite ("WAVE", sizeof (char), 4, pfTapeOut);
   fwrite ("fmt ", sizeof (char), 4, pfTapeOut);
   iVal = 18;	/* Chunk size */
   fwrite (&iVal, sizeof (iVal), 1, pfTapeOut);
   wVal = 1;	/* Format */
   fwrite (&wVal, sizeof (wVal), 1, pfTapeOut);
   wVal = 1;	/* Number of channels */
   fwrite (&wVal, sizeof (wVal), 1, pfTapeOut);
   fwrite (&nRate, sizeof (nRate), 1, pfTapeOut);
   fwrite (&nRate, sizeof (nRate), 1, pfTapeOut);
   wVal = 1;	/* Block align */
   fwrite (&wVal, sizeof (wVal), 1, pfTapeOut);
   iVal = 8;	/* Bits per sample */
   fwrite (&iVal, sizeof (iVal), 1, pfTapeOut);
   fwrite ("data", sizeof (char), 4, pfTapeOut);
   fwrite (&nSampOut, sizeof (nSampOut), 1, pfTapeOut);
   unsigned int nOne = nRate / nBaud + 1;
   unsigned int nZero = nRate / nBaud / 2;
   // printf ("Zero = %d, One = %d\n", nZero, nOne);
   for ( unsigned int iPulse = 0; iPulse < pulses.size (); ++iPulse )
      {
      unsigned int nDurn = nZero;
      if ( ( ( pulses[iPulse].flags & M_TYPE ) == P_ONE )
         || ( ( pulses[iPulse].flags & M_TYPE ) == P_LONG ) ) nDurn = nOne;
      unsigned char by = ( pulses[iPulse].lvl == lvlHigh ) ? 255 : 0;
      for ( unsigned int iDurn = 0; iDurn < nDurn; ++iDurn )
         {
         fwrite (&by, sizeof (by), 1, pfTapeOut);
         }
      nSampOut += nDurn;
      }
   // printf ("Wrote %d values.\n", nSampOut);
   /* Size of data chunk */
   fseek (pfTapeOut, 42, SEEK_SET);
   fwrite (&nSampOut, sizeof (nSampOut), 1, pfTapeOut);
   /* Size of WAVE file */
   nSampOut += 34;
   fseek (pfTapeOut, 4, SEEK_SET);
   fwrite (&nSampOut, sizeof (nSampOut), 1, pfTapeOut);
   fseek (pfTapeOut, 0, SEEK_END);
   // printf ("Updated file sizes.\n");
   fclose (pfTapeOut);
   // printf ("Closed file.\n");
   }

void Tape::SaveMTX (const char *psFile)
   {
   // printf ("Open file %s\n", psFile);
   FILE * pfMTXOut = fopen (psFile, "wb");
   unsigned int iBlock;
   for ( iBlock = 0; iBlock < nBlock; ++iBlock )
      {
      Block &block = blocks[iBlock];
      if ( block.fmt != fmtUnknown )
         {
         fwrite (&block.data[0], 1, block.data.size (), pfMTXOut);
         }
      }
   fclose (pfMTXOut);
   }

void Tape::ListCode (const char *psFile)
   {
   FILE * pfList = fopen (psFile, "w");
   wxString sULine;
   sULine.Append ('=', sSource.Len ());
   fprintf (pfList,
      "TapeView Analysis of file %s\n"
      "==========================%s\n"
      "Sample Rate: %d\n"
      "Baud Rate: %d\n"
      "Number of Samples: %d\n"
      "Analysis method: %s\n"
      "Thresholds, Up: %6.3f, Down: %6.3f\n"
      "Minimum leader: %d\n"
      "Pulse thresholds, Short: %d, Mid: %d, Long %d\n\n",
      (const char *) sSource, (const char *) sULine,
      nRate, nBaud, nSamp,
      (const char *) sScanMode[scn],
      fUp, fDn, nMinLead,
      Pulse::nShort, Pulse::nMid, Pulse::nLong);
   unsigned int iBlock;
   for ( iBlock = 0; iBlock < nBlock; ++iBlock )
      {
      fprintf (pfList, "Block %3d\n---------\n"
         "Format = %s Address = 0x%04X Length = 0x%04X\n\n",
         iBlock, (const char *) Block::sBlockFmt[blocks[iBlock].fmt],
         blocks[iBlock].nAddr, blocks[iBlock].data.size ());
      blocks[iBlock].ListCode ();
      fprintf (pfList, "%s\n", (const char *) blocks[iBlock].sList);
      }
   fclose (pfList);
   }

void Tape::SaveManual (const char *psFile)
   {
   FILE * pfSave = fopen (psFile, "w");
   fprintf (pfSave,
      "%s\n"
      "%s\n"
      "%6.3f\n%6.3f\n%d\n"
      "%d\n%d\n%d\n",
      (const char *) sSource,
      (const char *) sScanMode[scn],
      fUp, fDn, nMinLead,
      Pulse::nShort, Pulse::nMid, Pulse::nLong);
   for (std::set<unsigned int>::iterator it = split.begin (); it != split.end (); ++it)
      {
      fprintf (pfSave, "S%d\n", *it);
      }
   for (std::set<unsigned int>::iterator it = merge.begin (); it != merge.end (); ++it)
      {
      fprintf (pfSave, "M%d\n", *it);
      }
   fclose (pfSave);
   }

bool Tape::LoadManual (const char *psFile)
   {
   char sLine[512];
   FILE *pfLoad = fopen (psFile, "r");
   fgets (sLine, sizeof (sLine), pfLoad);
   sLine[strlen (sLine) - 1] = '\0';
   if ( sSource != sLine )
      {
      fclose (pfLoad);
      return false;
      }
   fgets (sLine, sizeof (sLine), pfLoad);
   sLine[strlen (sLine) - 1] = '\0';
   ScanMode scnIn = scnValue;
   for (int iMode = 0; iMode < scnCount; ++iMode)
      {
      if ( sScanMode[iMode] == sLine )
         {
         scnIn = (enum ScanMode) iMode;
         break;
         }
      }
   if ( ! bLockPulses ) scn = scnIn;
   fgets (sLine, sizeof (sLine), pfLoad);
   if ( ! bLockPulses ) fUp = atof (sLine);
   fgets (sLine, sizeof (sLine), pfLoad);
   if ( ! bLockPulses ) fDn = atof (sLine);
   fgets (sLine, sizeof (sLine), pfLoad);
   if ( ! bLockPulses ) nMinLead = atoi (sLine);
   fgets (sLine, sizeof (sLine), pfLoad);
   if ( ! bLockPulses ) Pulse::nShort = atoi (sLine);
   fgets (sLine, sizeof (sLine), pfLoad);
   if ( ! bLockPulses ) Pulse::nMid = atoi (sLine);
   fgets (sLine, sizeof (sLine), pfLoad);
   if ( ! bLockPulses ) Pulse::nLong = atoi (sLine);
   split.clear ();
   merge.clear ();
   while ( fgets (sLine, sizeof (sLine), pfLoad) != NULL )
      {
      unsigned int iSamp = atoi (&sLine[1]);
      if ( sLine[0] == 'S' ) split.insert (iSamp);
      else if ( sLine[0] == 'M' ) merge.insert (iSamp);
      }
   fclose (pfLoad);
   return true;
   }

void Tape::ApplyManual (void)
   {
   std::set<unsigned int>::iterator it;
   unsigned int iPulse = 0;
   for ( it = split.begin (); it != split.end (); ++it )
      {
      iPulse = SplitPulse (iPulse, *it);
      }
   iPulse = 0;
   for ( it = merge.begin (); it != merge.end (); ++it )
      {
      iPulse = MergePulse (iPulse, *it);
      }
   }

unsigned int Tape::SysWord (unsigned int iBlock, unsigned int iAddr)
    {
    unsigned int w = 0;
    while ( --iBlock >= 0 )
        {
        if ( blocks[iBlock].fmt == fmtSysVar )
            {
            w = blocks[iBlock].Word (iAddr);
            break;
            }
        }
    return w;
    }

class TapeView: public wxApp
   {
public:
   virtual bool OnInit ();
   };

class TVFrame: public wxFrame
   {
public:
   TVFrame (const wxString& title, const wxPoint& pos, const wxSize& size);
private:
   Tape tape;
   wxWindow * wdChart;
   wxButton * btBackBad;
   wxButton * btBackBlock;
   wxButton * btBackMark;
   wxButton * btBackPulse;
   wxButton * btNextPulse;
   wxButton * btNextMark;
   wxButton * btNextBlock;
   wxButton * btNextBad;
   wxButton * btZoomIn;
   wxButton * btZoomOut;
   wxButton * btZoomOne;
   wxButton * btZoomFit;
   wxTextCtrl * txUpCross;
   wxTextCtrl * txDnCross;
   wxTextCtrl * txMinLead;
   wxTextCtrl * txShort;
   wxTextCtrl * txMid;
   wxTextCtrl * txLong;
   wxComboBox * cbScanMode;
   wxButton * btScan;
   wxSpinCtrl * spNBlock;
   wxComboBox * cbBlkFmt;
   wxTextCtrl * txBlkAddr;
   wxTextCtrl * txBlkLen;
   wxTextCtrl * txBlock;
   wxTextCtrl * txByteAddr;
   wxButton * btByteGoto;
   wxBoxSizer * bsz1;
   wxBoxSizer * bsz2;
   wxFlexGridSizer * fsz3;
//   wxBoxSizer * bsz4;
   wxPen pens[4];
   wxPen penGrey;
   wxBrush brYellow;
   void TapeStatus (void);
   void DoScan (void);
   void ShowBlock (unsigned int iBlock);
   void OnOpen (wxCommandEvent& event);
   void OnExit (wxCommandEvent& event);
   void OnAbout (wxCommandEvent& event);
   void OnMove (wxCommandEvent& event);
   void OnScan (wxCommandEvent& event);
   void OnPaint (wxPaintEvent& event);
   void OnMouse (wxMouseEvent& event);
   void OnBlockChg (wxSpinEvent& event);
   void OnBlockFmt (wxCommandEvent& event);
   void OnBlockAddr (wxCommandEvent& evevt);
   void OnBlockLen (wxCommandEvent& event);
   void OnByteGoto (wxCommandEvent& event);
   void OnSaveWave (wxCommandEvent& event);
   void OnSaveMTX (wxCommandEvent& event);
   void OnListCode (wxCommandEvent& event);
   void OnLockPulses (wxCommandEvent& event);
   void OnManualLoad (wxCommandEvent& event);
   void OnManualSave (wxCommandEvent& event);
   wxDECLARE_EVENT_TABLE();
   };

enum
   {
   ID_BACK_BAD = 1,
   ID_BACK_BLOCK,
   ID_BACK_MARK,
   ID_BACK_PULSE,
   ID_NEXT_PULSE,
   ID_NEXT_MARK,
   ID_NEXT_BLOCK,
   ID_NEXT_BAD,
   ID_ZOOM_IN,
   ID_ZOOM_OUT,
   ID_ZOOM_ONE,
   ID_ZOOM_FIT,
   ID_UP_CROSS,
   ID_DN_CROSS,
   ID_MIN_LEAD,
   ID_SHORT_PULSE,
   ID_MID_PULSE,
   ID_LONG_PULSE,
   ID_SCAN_MODE,
   ID_SCAN,
   ID_BLOCK_CHOOSE,
   ID_BLOCK_FMT,
   ID_BLOCK_ADDR,
   ID_BLOCK_LEN,
   ID_BLOCK_SHOW,
   ID_BYTE_ADDR,
   ID_BYTE_GOTO,
   ID_SAVE_WAVE,
   ID_SAVE_MTX,
   ID_LIST_CODE,
   ID_LOCK_PULSES,
   ID_MANUAL_LOAD,
   ID_MANUAL_SAVE
   };

wxBEGIN_EVENT_TABLE(TVFrame, wxFrame)
EVT_MENU(wxID_OPEN,  TVFrame::OnOpen)
EVT_MENU(wxID_EXIT,  TVFrame::OnExit)
EVT_MENU(wxID_ABOUT,  TVFrame::OnAbout)
EVT_MENU(ID_SAVE_WAVE, TVFrame::OnSaveWave)
EVT_MENU(ID_SAVE_MTX, TVFrame::OnSaveMTX)
EVT_MENU(ID_LIST_CODE, TVFrame::OnListCode)
EVT_MENU(ID_LOCK_PULSES, TVFrame::OnLockPulses)
EVT_MENU(ID_MANUAL_LOAD, TVFrame::OnManualLoad)
EVT_MENU(ID_MANUAL_SAVE, TVFrame::OnManualSave)
EVT_BUTTON(ID_BACK_BAD, TVFrame::OnMove)
EVT_BUTTON(ID_BACK_BLOCK, TVFrame::OnMove)
EVT_BUTTON(ID_BACK_MARK, TVFrame::OnMove)
EVT_BUTTON(ID_BACK_PULSE, TVFrame::OnMove)
EVT_BUTTON(ID_NEXT_PULSE, TVFrame::OnMove)
EVT_BUTTON(ID_NEXT_MARK, TVFrame::OnMove)
EVT_BUTTON(ID_NEXT_BLOCK, TVFrame::OnMove)
EVT_BUTTON(ID_NEXT_BAD, TVFrame::OnMove)
EVT_BUTTON(ID_ZOOM_IN, TVFrame::OnMove)
EVT_BUTTON(ID_ZOOM_OUT, TVFrame::OnMove)
EVT_BUTTON(ID_ZOOM_ONE, TVFrame::OnMove)
EVT_BUTTON(ID_SCAN, TVFrame::OnScan)
EVT_SPINCTRL(ID_BLOCK_CHOOSE, TVFrame::OnBlockChg)
EVT_COMBOBOX(ID_BLOCK_FMT, TVFrame::OnBlockFmt)
EVT_TEXT(ID_BLOCK_ADDR, TVFrame::OnBlockAddr)
EVT_TEXT(ID_BLOCK_LEN, TVFrame::OnBlockLen)
EVT_BUTTON(ID_BYTE_GOTO, TVFrame::OnByteGoto)
wxEND_EVENT_TABLE()

wxIMPLEMENT_APP(TapeView);

bool TapeView::OnInit ()
   {
   TVFrame *frame = new TVFrame ("TapeView", wxDefaultPosition, wxDefaultSize);
   frame->Show (true);
   return true;
   }

TVFrame::TVFrame (const wxString& title, const wxPoint& pos, const wxSize& size)
   : wxFrame (NULL, wxID_ANY, title, pos, size)
   {
   wdChart = new wxWindow (this, wxID_ANY);
   btBackBad = new wxButton (this, ID_BACK_BAD, "Back Bad");
   btBackBlock = new wxButton (this, ID_BACK_BLOCK, "Back Block");
   btBackMark = new wxButton (this, ID_BACK_MARK, "Back Mark");
   btBackPulse = new wxButton (this, ID_BACK_PULSE, "Back Pulse");
   btNextPulse = new wxButton (this, ID_NEXT_PULSE, "Next Pulse");
   btNextMark = new wxButton (this, ID_NEXT_MARK, "Next Mark");
   btNextBlock = new wxButton (this, ID_NEXT_BLOCK, "Next Block");
   btNextBad = new wxButton (this, ID_NEXT_BAD, "Next Bad");
   btZoomIn = new wxButton (this, ID_ZOOM_IN, "Zoom In");
   btZoomOut = new wxButton (this, ID_ZOOM_OUT, "Zoom Out");
   btZoomOne = new wxButton (this, ID_ZOOM_ONE, "Zoom One");
   btZoomFit = new wxButton (this, ID_ZOOM_FIT, "Zoom Fit");
   wxFloatingPointValidator<float> vlUpCross (4, &tape.fUp, wxNUM_VAL_NO_TRAILING_ZEROES);
   wxFloatingPointValidator<float> vlDnCross (4, &tape.fDn, wxNUM_VAL_NO_TRAILING_ZEROES);
   wxIntegerValidator<unsigned int> vlMinLead (&tape.nMinLead);
   wxIntegerValidator<unsigned int> vlShort (&Pulse::nShort);
   wxIntegerValidator<unsigned int> vlMid (&Pulse::nMid);
   wxIntegerValidator<unsigned int> vlLong (&Pulse::nLong);
   wxGenericValidator vlScanMode ((int *) &tape.scn);
   vlUpCross.SetRange (-1.0, 1.0);
   vlDnCross.SetRange (-1.0, 1.0);
   vlMinLead.SetMin (1);
   vlShort.SetMin (1);
   vlMid.SetMin (2);
   vlLong.SetMin (3);
   txUpCross = new wxTextCtrl (this, ID_UP_CROSS, wxEmptyString,
      wxDefaultPosition, wxDefaultSize, 0, vlUpCross);
   txDnCross = new wxTextCtrl (this, ID_DN_CROSS, wxEmptyString,
      wxDefaultPosition, wxDefaultSize, 0, vlDnCross);
   txMinLead = new wxTextCtrl (this, ID_MIN_LEAD, wxEmptyString,
      wxDefaultPosition, wxDefaultSize, 0, vlMinLead);
   txShort = new wxTextCtrl (this, ID_SHORT_PULSE, wxEmptyString,
      wxDefaultPosition, wxDefaultSize, 0, vlShort);
   txMid = new wxTextCtrl (this, ID_MID_PULSE, wxEmptyString,
      wxDefaultPosition, wxDefaultSize, 0, vlMid);
   txLong = new wxTextCtrl (this, ID_LONG_PULSE, wxEmptyString,
      wxDefaultPosition, wxDefaultSize, 0, vlLong);
   cbScanMode = new wxComboBox (this, ID_SCAN_MODE, wxEmptyString,
      wxDefaultPosition, wxDefaultSize,
      sizeof (Tape::sScanMode) / sizeof (Tape::sScanMode[0]), Tape::sScanMode,
      wxCB_DROPDOWN | wxCB_READONLY, vlScanMode);
   btScan = new wxButton (this, ID_SCAN, "Scan");
//   cbBlock = new wxComboBox (this, ID_BLOCK_CHOOSE, wxEmptyString,
//      wxDefaultPosition, wxDefaultSize, 0, NULL,
//      wxCB_DROPDOWN | wxCB_READONLY);
   spNBlock = new wxSpinCtrl (this, ID_BLOCK_CHOOSE);
   cbBlkFmt = new wxComboBox (this, ID_BLOCK_FMT, wxEmptyString,
      wxDefaultPosition, wxDefaultSize,
      sizeof (Block::sBlockFmt) / sizeof (Block::sBlockFmt[0]),
      Block::sBlockFmt, wxCB_DROPDOWN | wxCB_READONLY);
   txBlkAddr = new wxTextCtrl (this, ID_BLOCK_ADDR);
   txBlkLen = new wxTextCtrl (this, ID_BLOCK_LEN);
   txByteAddr = new wxTextCtrl (this, ID_BYTE_ADDR);
   btByteGoto = new wxButton (this, ID_BYTE_GOTO, "Goto Byte");
   txBlock = new wxTextCtrl (this, ID_BLOCK_SHOW, wxEmptyString,
      wxDefaultPosition, wxDefaultSize,
      wxTE_MULTILINE | wxTE_READONLY | wxHSCROLL);
   wxTextAttr ta;
   ta.SetFontFamily (wxFONTFAMILY_TELETYPE);
   txBlock->SetDefaultStyle (ta);
   bsz1 = new wxBoxSizer (wxVERTICAL);
   bsz2 = new wxBoxSizer (wxHORIZONTAL);
   fsz3 = new wxFlexGridSizer (8);
   fsz3->AddGrowableCol (0);
   fsz3->AddGrowableCol (2);
   fsz3->AddGrowableCol (4);
   fsz3->AddGrowableCol (6);
   fsz3->SetFlexibleDirection (wxHORIZONTAL);
//   bsz4 = new wxBoxSizer (wxHORIZONTAL);
   wxMenu *menuFile = new wxMenu;
   menuFile->Append (wxID_OPEN);
   menuFile->Append (ID_SAVE_WAVE, "Save &Wave", "Create new .wav file with normalised pulses");
   menuFile->Append (ID_SAVE_MTX, "Save &MTX", "Save decoded bytes as .mtx file for loading into MEMU");
   menuFile->Append (ID_LIST_CODE, "&List Code", "Create a text file reporting analysis and bytes decoded");
   menuFile->Append (wxID_EXIT);
   wxMenu *menuManual = new wxMenu;
   menuManual->Append (ID_LOCK_PULSES, "Lock &Pulses",
      "Scan does not change pulse definitions (for manual editing)", wxITEM_CHECK);
   menuManual->Append (ID_MANUAL_LOAD, "&Load Manual", "Load manual pulse definitions from file");
   menuManual->Append (ID_MANUAL_SAVE, "&Save Manual", "Save manual pulse definitions to file");
   wxMenu *menuHelp = new wxMenu;
   menuHelp->Append (wxID_ABOUT);
   wxMenuBar *menuBar = new wxMenuBar;
   menuBar->Append (menuFile, "&File");
   menuBar->Append (menuManual, "&Manual");
   menuBar->Append (menuHelp, "&Help");
   SetMenuBar (menuBar);
   CreateStatusBar ();
   TapeStatus ();
   bsz2->Add (btBackBad, 1, wxALIGN_CENTRE);
   bsz2->Add (btBackBlock, 1, wxALIGN_CENTRE);
   bsz2->Add (btBackMark, 1, wxALIGN_CENTRE);
   bsz2->Add (btBackPulse, 1, wxALIGN_CENTRE);
   bsz2->Add (btNextPulse, 1, wxALIGN_CENTRE);
   bsz2->Add (btNextMark, 1, wxALIGN_CENTRE);
   bsz2->Add (btNextBlock, 1, wxALIGN_CENTRE);
   bsz2->Add (btNextBad, 1, wxALIGN_CENTRE);
   bsz2->Add (btZoomIn, 1, wxALIGN_CENTRE);
   bsz2->Add (btZoomOut, 1, wxALIGN_CENTRE);
   bsz2->Add (btZoomOne, 1, wxALIGN_CENTRE);
   bsz2->Add (btZoomFit, 1, wxALIGN_CENTRE);
   fsz3->Add (new wxStaticText (this, wxID_ANY, "Up Cross:",
            wxDefaultPosition, wxDefaultSize, wxALIGN_RIGHT), 1,
      wxALIGN_RIGHT | wxALIGN_CENTRE_VERTICAL);
   fsz3->Add (txUpCross, 1, wxEXPAND);
   fsz3->Add (new wxStaticText (this, wxID_ANY, "Down cross:",
            wxDefaultPosition, wxDefaultSize, wxALIGN_RIGHT), 1,
      wxALIGN_RIGHT | wxALIGN_CENTRE_VERTICAL);
   fsz3->Add (txDnCross, 1, wxEXPAND);
   fsz3->Add (new wxStaticText (this, wxID_ANY, "Min Lead:",
            wxDefaultPosition, wxDefaultSize, wxALIGN_RIGHT), 1,
      wxALIGN_RIGHT | wxALIGN_CENTRE_VERTICAL);
   fsz3->Add (txMinLead, 1, wxEXPAND);
   fsz3->Add (new wxStaticText (this, wxID_ANY, "Scan Mode:",
            wxDefaultPosition, wxDefaultSize, wxALIGN_RIGHT), 1,
      wxALIGN_RIGHT | wxALIGN_CENTRE_VERTICAL);
   fsz3->Add (cbScanMode, 1, wxEXPAND);
   // fsz3->AddSpacer (0);
   // fsz3->AddSpacer (0);
   fsz3->Add (new wxStaticText (this, wxID_ANY, "Short Pulse:",
            wxDefaultPosition, wxDefaultSize, wxALIGN_RIGHT), 1,
      wxALIGN_RIGHT | wxALIGN_CENTRE_VERTICAL);
   fsz3->Add (txShort, 1, wxEXPAND);
   fsz3->Add (new wxStaticText (this, wxID_ANY, "Mid Pulse:",
            wxDefaultPosition, wxDefaultSize, wxALIGN_RIGHT), 1,
      wxALIGN_RIGHT | wxALIGN_CENTRE_VERTICAL);
   fsz3->Add (txMid, 1, wxEXPAND);
   fsz3->Add (new wxStaticText (this, wxID_ANY, "Long Pulse:",
            wxDefaultPosition, wxDefaultSize, wxALIGN_RIGHT), 1,
      wxALIGN_RIGHT | wxALIGN_CENTRE_VERTICAL);
   fsz3->Add (txLong, 1, wxEXPAND);
   fsz3->AddSpacer (0);
   fsz3->Add (btScan, 1, wxALIGN_RIGHT);
   fsz3->Add (new wxStaticText (this, wxID_ANY, "Block Number:",
            wxDefaultPosition, wxDefaultSize, wxALIGN_RIGHT), 1,
      wxALIGN_RIGHT | wxALIGN_CENTRE_VERTICAL);
   fsz3->Add (spNBlock, 0, wxEXPAND);
   fsz3->Add (new wxStaticText (this, wxID_ANY, "Format:",
            wxDefaultPosition, wxDefaultSize, wxALIGN_RIGHT), 1,
      wxALIGN_RIGHT | wxALIGN_CENTRE_VERTICAL);
   fsz3->Add (cbBlkFmt, 0, wxEXPAND);
   fsz3->Add (new wxStaticText (this, wxID_ANY, "Start Address:",
            wxDefaultPosition, wxDefaultSize, wxALIGN_RIGHT), 1,
      wxALIGN_RIGHT | wxALIGN_CENTRE_VERTICAL);
   fsz3->Add (txBlkAddr, 0, wxEXPAND);
   fsz3->Add (new wxStaticText (this, wxID_ANY, "Block Length:",
            wxDefaultPosition, wxDefaultSize, wxALIGN_RIGHT), 1,
      wxALIGN_RIGHT | wxALIGN_CENTRE_VERTICAL);
   fsz3->Add (txBlkLen, 0, wxEXPAND);
   fsz3->Add (new wxStaticText (this, wxID_ANY, "Byte Addr:",
            wxDefaultPosition, wxDefaultSize, wxALIGN_RIGHT), 1,
      wxALIGN_RIGHT | wxALIGN_CENTRE_VERTICAL);
   fsz3->Add (txByteAddr, 0, wxEXPAND);
   fsz3->Add (btByteGoto, 0, wxALIGN_LEFT);
   bsz1->Add (wdChart, 1, wxEXPAND);
   bsz1->Add (bsz2, 0, wxEXPAND);
   bsz1->Add (fsz3, 0, wxEXPAND);
//   bsz1->Add (bsz4, 0, wxEXPAND);
   bsz1->Add (txBlock, 1, wxEXPAND);
   SetSizer (bsz1);
   pens[P_ZERO] = wxPen (wxColour (0x00U, 0x00U, 0x00U));
   pens[P_ONE] = wxPen (wxColour (0x00U, 0x00U, 0xFFU));
   pens[P_SHORT] = wxPen (wxColour (0xFFU, 0x00U, 0x00U));
   pens[P_LONG] = wxPen (wxColour (0xFFU, 0x99U, 0x00U));
   penGrey = wxPen (wxColour (0x99U, 0x99U, 0x99U));
   brYellow = wxBrush (wxColour (0xFFU, 0xFFU, 0x00U));
   wdChart->Bind (wxEVT_PAINT, &TVFrame::OnPaint, this);
   wdChart->Bind (wxEVT_LEFT_DOWN, &TVFrame::OnMouse, this);
//   wdChart->Bind (wxEVT_LEFT_UP, &TVFrame::OnMouse, this);
   wdChart->Bind (wxEVT_RIGHT_DOWN, &TVFrame::OnMouse, this);
   TransferDataToWindow ();
   Maximize (true);
   }

void TVFrame::OnExit (wxCommandEvent& event)
   {
   Close (true);
   }

void TVFrame::OnAbout (wxCommandEvent& event)
   {
   wxMessageBox ("Program for recovering data from MTX Tapes v1.2\nBill Brendling, 2019",
      "About TapeView", wxOK | wxICON_INFORMATION);
   }

void TVFrame::OnOpen (wxCommandEvent& event)
   {
   wxFileDialog OpenFD (this, "Open tape audio file", "", "",
                       "WAV files (*.wav)|*.wav|All files (*.*)|*.*", wxFD_OPEN|wxFD_FILE_MUST_EXIST);
   if ( OpenFD.ShowModal() == wxID_OK )
      {
      wxBusyCursor wait;
      // printf ("File path = %s\n", (const char *) OpenFD.GetPath ());
      bool bStatus = tape.Open (OpenFD.GetPath ());
      // printf ("Open = %d\n", bStatus);
      if ( bStatus )
         {
         SetStatusText ("Loading " + tape.sSource);
         bStatus = tape.Load ();
         // printf ("Load = %d\n", bStatus);
         }
      if ( bStatus )
         {
         DoScan ();
         // printf ("Scan.\n");
         TapeStatus ();
         }
      else SetStatusText (tape.sMsg);
      // printf ("Done.\n");
      TransferDataToWindow ();
      }
   }

void TVFrame::OnPaint (wxPaintEvent& event)
   {
   wxPaintDC dc(wdChart);
   if ( tape.samples.size () )
      {
      wxBusyCursor wait;
      // printf ("Paint.\n");
      unsigned int iPulse = tape.nFstPulse;
      unsigned int nFstSamp = tape.pulses[iPulse].n0;
      wxRect rcSize = wdChart->GetClientRect ();
      // printf ("Size: x = %d, y = %d, w = %d, h = %d.\n",
      //   rcSize.x, rcSize.y, rcSize.width, rcSize.height);
      rcSize.x += 10;
      rcSize.width -= 20;
      int htZero1 = rcSize.height / 2;
      int htZero2 = htZero1;
      if ( ( tape.scn != scnValue ) && ( tape.scn != scnHiLow ) )
         {
         htZero1 /= 2;
         htZero2 = 3 * rcSize.height / 4;
         }
      int htPlt = htZero1 - 2;
      wxRect rcUpd = wdChart->GetUpdateClientRect ();
      // printf ("Update: x = %d, y = %d, w = %d, h = %d.\n",
      //   rcUpd.x, rcUpd.y, rcUpd.width, rcUpd.height);
      int nRight = rcUpd.GetRight ();
      int nRight2 = rcSize.GetRight ();
      if ( nRight2 < nRight ) nRight = nRight2;
      while ( int (tape.fZoom * ( tape.pulses[iPulse].n2 - nFstSamp )) < rcUpd.x ) ++iPulse;
      unsigned int iSamp = tape.pulses[iPulse].n0;
      // printf ("First pulse = %d %d, sample = %d.\n", tape.nFstPulse, iPulse, iSamp);
      wxPoint pt1;
      wxPoint pt2;
      wxPoint pt3;
      wxPoint pt4;
      if ( tape.bShowByte )
          {
          pt1.x = rcSize.x + int (tape.fZoom * ( tape.pulses[tape.nShowPulse].n1 - nFstSamp ));
          pt2.x = rcSize.x + int (tape.fZoom * ( tape.pulses[tape.nShowPulse+15].n2 - nFstSamp ));
          if ( ( pt1.x <= nRight ) && ( pt2.x >= rcUpd.x ) )
              {
              dc.SetPen (wxNullPen);
              dc.SetBrush (brYellow);
              dc.DrawRectangle (pt1.x, rcSize.y, pt2.x - pt1.x, rcSize.height);
              dc.SetBrush (wxNullBrush);
              }
          }
      pt1.x = rcSize.x + int (tape.fZoom * ( iSamp - nFstSamp ));
      pt1.y = htZero1 - int (htPlt * tape.samples[iSamp]);
      if ( ( tape.scn != scnValue ) && ( tape.scn != scnHiLow ) )
         {
         pt3.x = pt1.x;
         pt3.y = htZero2 - int (htPlt * tape.Test (tape.scn, iSamp));
         }
      dc.SetPen (pens[tape.pulses[iPulse].flags & M_TYPE]);
      while ( ++iSamp < tape.nSamp )
         {
         pt2.x = rcSize.x + int (tape.fZoom * ( iSamp - nFstSamp ));
         if ( pt2.x >= nRight ) break;
         if ( iSamp > tape.pulses[iPulse].n2 )
            {
            ++iPulse;
            dc.SetPen (pens[tape.pulses[iPulse].flags & M_TYPE]);
            // printf ("Pulse = %d, flags = 0x%02X, Pen = %d.\n", iPulse, tape.pulses[iPulse].flags,
            //   tape.pulses[iPulse].flags & M_TYPE);
            }
         pt2.y = htZero1 - int (htPlt * tape.samples[iSamp]);
         // printf ("(%d, %d) pulse = %d, sample = %d, value = %5.3f.\n", pt2.x, pt2.y, iPulse, iSamp,
         //   tape.samples[iSamp]);
         dc.DrawLine (pt1, pt2);
         pt1 = pt2;
         if ( ( tape.scn != scnValue ) && ( tape.scn != scnHiLow ) )
            {
            pt4.x = pt2.x;
            pt4.y = htZero2 - int (htPlt * tape.Test (tape.scn, iSamp));
            dc.DrawLine (pt3, pt4);
            pt3 = pt4;
            }
         }
      dc.SetPen (penGrey);
      if ( tape.scn < scnHiLow )
         {
         pt1.x = rcSize.x;
         pt2.x = rcSize.GetRight ();
         pt1.y = htZero2 - int (htPlt * tape.fUp);
         pt2.y = pt1.y;
         dc.DrawLine (pt1, pt2);
         pt1.y = htZero2 - int (htPlt * tape.fDn);
         pt2.y = pt1.y;
         dc.DrawLine (pt1, pt2);
         }
      pt1.x = rcSize.x + int (tape.fZoom * ( tape.pulses[tape.nFstPulse].n1 - nFstSamp ));
      pt2.x = pt1.x;
      pt1.y = 0;
      pt2.y = rcSize.height;
      dc.DrawLine (pt1, pt2);
      if ( tape.fZoom >= 1.0 )
         {
         iPulse = tape.nFstPulse + 1;
         while ( iPulse < tape.pulses.size () )
            {
            pt1.x = rcSize.x + int (tape.fZoom * ( tape.pulses[iPulse].n1 - nFstSamp ));
            if ( pt1.x > nRight ) break;
            pt2.x = pt1.x;
            dc.DrawLine (pt1, pt2);
            ++iPulse;
            }
         }
      }
   }

void TVFrame::TapeStatus (void)
   {
   char sMsg[256];
   char sMsg2[128];
   char sFlg[32];
   sFlg[0] = '\0';
   if ( tape.nSamp == 0 )
      {
      SetStatusText ("No tape loaded");
      return;
      }
   else if ( tape.nFstPulse < tape.pulses.size () )
      {
      switch (tape.pulses[tape.nFstPulse].flags & M_TYPE)
         {
         case P_SHORT: strcpy (sFlg, "Short"); break;
         case P_LONG: strcpy (sFlg, "Long"); break;
         case P_ZERO: strcpy (sFlg, "Zero"); break;
         case P_ONE: strcpy (sFlg, "One"); break;
         }
      switch (tape.pulses[tape.nFstPulse].flags & M_MARK)
         {
         case P_LEAD: strcat (sFlg, " Lead"); break;
         case P_MARK: strcat (sFlg, " Mark"); break;
         case P_DATA: strcat (sFlg, " Data"); break;
         }
      snprintf (sMsg, sizeof (sMsg), "Pulse %d/%d: %s (%d, %d, 0x%02X), Sample %d/%d, Zoom %10.3E",
         tape.nFstPulse,
         tape.pulses.size (),
         sFlg,
         tape.pulses[tape.nFstPulse].n1 - tape.pulses[tape.nFstPulse].n0,
         tape.pulses[tape.nFstPulse].n2 - tape.pulses[tape.nFstPulse].n1,
         tape.pulses[tape.nFstPulse].flags,
         tape.pulses[tape.nFstPulse].n1,
         tape.nSamp,
         tape.fZoom);
      bool bInBlock = false;
      for (unsigned int iBlock = 0; iBlock < tape.nBlock; ++iBlock)
         {
         if ( ( tape.nFstPulse >= tape.blocks[iBlock].nPulse )
            && ( tape.nFstPulse < tape.blocks[iBlock].nPulse + 16 * tape.blocks[iBlock].data.size () ) )
            {
            bInBlock = true;
            snprintf (sMsg2, sizeof (sMsg2), " Block %d/%d Byte 0x%04X Bit %d",
               iBlock + 1,
               tape.nBlock,
               tape.blocks[iBlock].nAddr + ( tape.nFstPulse - tape.blocks[iBlock].nPulse ) / 16,
               ( tape.nFstPulse - tape.blocks[iBlock].nPulse ) % 16 / 2);
            break;
            }
         }
      if ( ! bInBlock ) snprintf (sMsg2, sizeof (sMsg2), " Block -/%d", tape.nBlock);
      strcat (sMsg, sMsg2);
      }
   else
      {
      snprintf (sMsg, sizeof (sMsg), "Pulse %d/%d: %s, Sample ?/%d, Zoom %5.3f",
         tape.nFstPulse,
         tape.pulses.size (),
         sFlg,
         tape.nSamp,
         tape.fZoom);
      }
   SetStatusText (sMsg);
   }

void TVFrame::OnMove (wxCommandEvent& event)
   {
   wxBusyCursor wait;
   unsigned int iPulse = tape.nFstPulse;
   switch (event.GetId ())
      {
      case ID_BACK_BAD:
         {
         // printf ("Back Bad.\n");
         if ( iPulse > 0 ) --iPulse;
         while ( ( iPulse > 0 ) && ( tape.pulses[iPulse].flags & M_VALID ) ) --iPulse;
         tape.bShowByte = false;
         break;
         }
      case ID_BACK_BLOCK:
         {
         unsigned int iBlock = 0;
         while ( ( iBlock < tape.blocks.size () ) && ( tape.blocks[iBlock].nPulse < iPulse ) )
            ++iBlock;
         if ( iBlock > 0 )
            {
            iPulse = tape.blocks[--iBlock].nPulse;
            spNBlock->SetValue (iBlock + 1);
            ShowBlock (iBlock);
            }
         tape.bShowByte = false;
         break;
         }
      case ID_BACK_MARK:
         {
         // printf ("Back Mark.\n");
         if ( iPulse > 0 ) --iPulse;
         while ( ( iPulse > 0 ) && ( ! ( tape.pulses[iPulse].flags & M_MARK ) ) ) --iPulse;
         tape.bShowByte = false;
         break;
         }
      case ID_BACK_PULSE:
         {
         // printf ("Back Pulse.\n");
         if ( iPulse > 0 ) --iPulse;
         break;
         }
      case ID_NEXT_PULSE:
         {
         // printf ("NextPulse.\n");
         if ( iPulse < tape.pulses.size () - 1 ) ++iPulse;
         break;
         }
      case ID_NEXT_MARK:
         {
         // printf ("Next Mark: pulse = iPulse\n");
         unsigned int nMaxPulse = tape.pulses.size () - 1;
         if ( iPulse < nMaxPulse ) ++iPulse;
         while ( ( iPulse < nMaxPulse ) && ( ! ( tape.pulses[iPulse].flags & M_MARK ) ) ) ++iPulse;
         // printf ("New pulse = %d, flags = 0x%02X.\n", iPulse, tape.pulses[iPulse].flags);
         tape.bShowByte = false;
         break;
         }
      case ID_NEXT_BLOCK:
         {
         unsigned int iBlock = 0;
         while ( ( iBlock < tape.blocks.size () ) && ( tape.blocks[iBlock].nPulse <= iPulse ) )
            ++iBlock;
         if ( iBlock < tape.blocks.size () )
            {
            iPulse = tape.blocks[iBlock].nPulse;
            spNBlock->SetValue (iBlock + 1);
            ShowBlock (iBlock);
            }
         tape.bShowByte = false;
         break;
         }
      case ID_NEXT_BAD:
         {
         // printf ("Next Bad: pulse = iPulse\n");
         unsigned int nMaxPulse = tape.pulses.size () - 1;
         if ( iPulse < nMaxPulse ) ++iPulse;
         while ( ( iPulse < nMaxPulse ) && ( tape.pulses[iPulse].flags & M_VALID ) ) ++iPulse;
         // printf ("New pulse = %d, flags = 0x%02X.\n", iPulse, tape.pulses[iPulse].flags);
         tape.bShowByte = false;
         break;
         }
      case ID_ZOOM_IN:
         {
         // printf ("Zoom In.\n");
         tape.fZoom *= 2.0;
         break;
         }
      case ID_ZOOM_OUT:
         {
         // printf ("Zoom Out.\n");
         tape.fZoom /= 2.0;
         break;
         }
      case ID_ZOOM_ONE:
         {
         // printf ("Zoom One.\n");
         tape.fZoom = 1.0;
         break;
         }
      case ID_ZOOM_FIT:
         {
         // printf ("Zoom Fit.\n");
         wxRect rcSize = wdChart->GetClientSize ();
         while ( tape.fZoom * tape.nSamp > rcSize.width ) tape.fZoom /= 2.0;
         break;
         }
      }
   tape.nFstPulse = iPulse;
   TapeStatus ();
   wdChart->Refresh ();
   }

void TVFrame::ShowBlock (unsigned int iBlock)
   {
   wxBusyCursor wait;
   Block & block = tape.blocks[iBlock];
//   cbBlkFmt->SetSelection (int (block.fmt));
   cbBlkFmt->ChangeValue (cbBlkFmt->GetString (block.fmt));
   txBlkAddr->ChangeValue (wxString::Format ("0x%04X", block.nAddr));
   txBlkLen->ChangeValue (wxString::Format ("0x%04X", block.nLen));
   block.List (txBlock);
   }

void TVFrame::DoScan (void)
   {
   unsigned int nSample = tape.nFstPulse ? tape.pulses[tape.nFstPulse].n1 : 0;
   SetStatusText ("Scanning ...");
   // printf ("Scan (%d, %d, %d)\n", Pulse::nShort, Pulse::nMid, Pulse::nLong);
   tape.Scan ();
   unsigned int nMaxPulse = tape.pulses.size () - 1;
   tape.nFstPulse = 0;
   while ( ( tape.nFstPulse < nMaxPulse ) && ( tape.pulses[tape.nFstPulse].n2 < nSample ) )
      ++tape.nFstPulse;
   TapeStatus ();
   if ( tape.nBlock > 0 )
      {
      spNBlock->SetValue (1);
      spNBlock->SetRange (1, tape.nBlock);
      ShowBlock (0);
      }
   else
      {
      spNBlock->SetValue (1);
      spNBlock->SetRange (1, 1);
      txBlock->SetValue ("No blocks found.");
      }
   tape.bShowByte = false;
   wdChart->Refresh ();
   }

void TVFrame::OnScan (wxCommandEvent& event)
   {
   wxBusyCursor wait;
   if ( Validate () && TransferDataFromWindow () )
      {
      if ( tape.fUp < tape.fDn )
         {
         wxMessageBox ("Up Crossing should be larger than Down Crossing",
            "TapeView Error", wxOK | wxICON_ERROR);
         }
      else if ( ( Pulse::nMid <= Pulse::nShort ) || ( Pulse::nMid >= Pulse::nLong ) )
         {
         wxMessageBox ("Mid Pulse Length should be between Short and Long",
            "TapeView Error", wxOK | wxICON_ERROR);
         }
      else
         {
         DoScan ();
         }
      }
   tape.bShowByte = false;
   }

void TVFrame::OnMouse (wxMouseEvent& event)
   {
   if ( event.ButtonDown () )
      {
      if ( tape.fZoom >= 1.0 )
         {
         wxRect rcSize = wdChart->GetClientRect ();
         // printf ("Size: x = %d, y = %d, w = %d, h = %d.\n",
         //   rcSize.x, rcSize.y, rcSize.width, rcSize.height);
         rcSize.x += 10;
         rcSize.width -= 20;
         int iX = event.GetX () - rcSize.x;
         unsigned int iPulse = tape.nFstPulse;
         unsigned int iSamp = tape.pulses[iPulse].n0 + int (iX / tape.fZoom + 0.5);
         // printf ("nFstPulse = %d, n0 = %d, n1 = %d, n2 = %d\n",
         //    iPulse, tape.pulses[iPulse].n0, tape.pulses[iPulse].n1, tape.pulses[iPulse].n2);
         // printf ("fZoom = %3.1f, iX = %d, iSamp = %d\n", tape.fZoom, iX, iSamp);
         if ( event.LeftDown () )
            {
            if ( tape.merge.find (iSamp) != tape.merge.end () ) tape.merge.erase (iSamp);
            else tape.split.insert (iSamp);
            tape.SplitPulse (iPulse, iSamp);
            TapeStatus ();
            wdChart->Refresh ();
            }
         else if ( event.RightDown () )
            {
            tape.MergePulse (iPulse, iSamp);
            if ( tape.split.find (iSamp) != tape.split.end () ) tape.split.erase (iSamp);
            else tape.merge.insert (iSamp);
            TapeStatus ();
            wdChart->Refresh ();
            }
         }
      }
   /*
   static int nXDrop;
   if ( event.ButtonDown () )
      {
      nXDrop = event.GetX ();
      }
   else
      {
      unsigned int nSample = tape.pulses[tape.nFstPulse].n1 + int (nXDrop / tape.fZoom);
      unsigned int nMaxPulse = tape.pulses.size () - 1;
      while ( ( tape.nFstPulse < nMaxPulse ) && ( tape.pulses[tape.nFstPulse].n2 < nSample ) )
         ++tape.nFstPulse;
      int nWide = event.GetX () - nXDrop;
      if ( nWide > 10 )
         {
         wxRect rcSize = wdChart->GetClientRect ();
         while ( nWide < rcSize.width / 2 )
            {
            nWide *= 2;
            tape.fZoom *= 2.0;
            }
         }
      }
   TapeStatus ();
   wdChart->Refresh ();
   */
   }

void TVFrame::OnBlockChg (wxSpinEvent& event)
   {
   wxBusyCursor wait;
   ShowBlock (event.GetPosition () - 1);
   }

void TVFrame::OnBlockFmt (wxCommandEvent& event)
   {
   unsigned int iBlock = spNBlock->GetValue () - 1;
   enum BlockFmt fmt = (enum BlockFmt) cbBlkFmt->GetSelection ();
   if ( tape.blocks[iBlock].fmt != fmt )
      {
      wxBusyCursor wait;
      tape.blocks[iBlock].fmt = fmt;
      tape.blocks[iBlock].sList = "";
      tape.blocks[iBlock].List (txBlock);
      }
   }

void TVFrame::OnBlockAddr (wxCommandEvent& event)
   {
   unsigned long int iVal;
   if ( txBlkAddr->GetValue ().ToCULong (&iVal, 0) )
      {
      unsigned int iBlock = spNBlock->GetValue () - 1;
      tape.blocks[iBlock].nAddr = iVal;
      }
   }

void TVFrame::OnBlockLen (wxCommandEvent& event)
   {
   unsigned long int iVal;
   if ( txBlkLen->GetValue ().ToCULong (&iVal, 0) )
      {
      unsigned int iBlock = spNBlock->GetValue () - 1;
      tape.blocks[iBlock].nLen = iVal;
      }
   }

void TVFrame::OnByteGoto (wxCommandEvent& event)
   {
   unsigned long int iAddr;
   if ( txByteAddr->GetValue ().ToCULong (&iAddr, 0) )
      {
      unsigned int iBlock = spNBlock->GetValue () - 1;
      Block & block = tape.blocks[iBlock];
      if ( iAddr >= block.nAddr )
         {
         iAddr -= block.nAddr;
         if ( iAddr < block.data.size () )
            {
            wxBusyCursor wait;
            tape.nFstPulse = block.nPulse + 16 * iAddr;
            tape.nShowPulse = tape.nFstPulse;
            tape.bShowByte = true;
            TapeStatus ();
            wdChart->Refresh ();
            }
         }
      }
   }

void TVFrame::OnSaveWave (wxCommandEvent &event)
   {
   wxString sDest = tape.sSource.BeforeLast ('.') + "_TV.wav";
   wxFileDialog fdSave (this, "Save WAV file", "", sDest, "WAV files (*.wav)|*.wav",
      wxFD_SAVE | wxFD_OVERWRITE_PROMPT);
   if ( fdSave.ShowModal () == wxID_OK )
      {
      wxBusyCursor wait;
      wxString sPath = fdSave.GetPath ();
      if ( ! sPath.EndsWith (".wav") ) sPath += ".wav";
      tape.SaveWave (sPath);
      }
   }

void TVFrame::OnSaveMTX (wxCommandEvent &event)
   {
   wxString sDest = tape.sSource.BeforeLast ('.') + ".mtx";
   wxFileDialog fdSave (this, "Save MTX file", "", sDest, "MTX files (*.mtx)|*.mtx",
      wxFD_SAVE | wxFD_OVERWRITE_PROMPT);
   if ( fdSave.ShowModal () == wxID_OK )
      {
      wxBusyCursor wait;
      wxString sPath = fdSave.GetPath ();
      if ( ! sPath.EndsWith (".mtx") ) sPath += ".mtx";
      tape.SaveMTX (sPath);
      }
   }

void TVFrame::OnListCode (wxCommandEvent &event)
   {
   wxString sDest = tape.sSource.BeforeLast ('.') + ".txt";
   wxFileDialog fdSave (this, "List Code Blocks", "", "", "Text files (*.txt)|*.txt",
      wxFD_SAVE | wxFD_OVERWRITE_PROMPT);
   if ( fdSave.ShowModal () == wxID_OK )
      {
      wxBusyCursor wait;
      wxString sPath = fdSave.GetPath ();
      if ( ! sPath.EndsWith (".txt") ) sPath += ".txt";
      tape.ListCode (sPath);
      }
   }

void TVFrame::OnLockPulses (wxCommandEvent& event)
   {
   tape.bLockPulses = event.IsChecked ();
   }

void TVFrame::OnManualLoad (wxCommandEvent& event)
   {
   wxFileDialog fdLoad (this, "Load Manual Pulse Definitions", "", "",
                       "Manual pulse files (*.mpd)|*.mpd", wxFD_OPEN|wxFD_FILE_MUST_EXIST);
   if ( fdLoad.ShowModal() == wxID_OK )
      {
      wxBusyCursor wait;
      if ( ! tape.LoadManual (fdLoad.GetPath ()) )
         {
         wxMessageBox ("Manual Pulse Definition file not appicable to this WAV file!",
            "TapeView Error", wxOK | wxICON_ERROR);
         return;
         }
      if ( ! tape.bLockPulses )
         {
         TransferDataToWindow ();
         tape.Scan ();
         }
      tape.ApplyManual ();
      }
   }

void TVFrame::OnManualSave (wxCommandEvent &event)
   {
   wxString sDest = tape.sSource.BeforeLast ('.') + ".mpd";
   wxFileDialog fdSave (this, "Save Manual Pulse Definitions", "", "",
      "Manual pulse files (*.mpd)|*.mpd", wxFD_SAVE | wxFD_OVERWRITE_PROMPT);
   if ( fdSave.ShowModal () == wxID_OK )
      {
      wxBusyCursor wait;
      wxString sPath = fdSave.GetPath ();
      if ( ! sPath.EndsWith (".mpd") ) sPath += ".mpd";
      tape.SaveManual (sPath);
      }
   }
